/* lif_phy.c -- interface to physical LIF disks */
/* 2000 A. R. Duell, and placed under the GPL */

/* This file contains machine-dependant functions to access sectors on
   a real LIF disk inserted in a PC floppy drive. This versions is for
   Linux */

/* It appears that an HP LIF disk (as used on the HP71/9114, etc) consists
   of 77 cylinders (numbered 0-76), 2 sides (0 and 1) and 16 sectors/track
   (1-16), 256 bytes/sector. Double density at the standard (256 kbps) data
   rate */

#include<stdio.h>
#include<unistd.h>
#include<sys/ioctl.h>
#include<linux/fd.h>
#include<linux/fdreg.h>

/* Data rate selection code for 250kbps */
#define RATE250 2

/* Read/write a single double density sector functions for the 8272 
   disk controller. These are not in fdreg.h */
#define FD_DD_WRITE 0x45
#define FD_DD_READ 0x46

/* Sector size in bytes */
#define SECTOR_SIZE 256

void lif_recalibrate(int device)
  {
    struct floppy_raw_cmd cmd;

    cmd.data=NULL;  /* pointer to data buffer */
    cmd.length=0;   /* length of DMA transfer */
    cmd.rate=RATE250; /* data rate = 250kbps */
    cmd.flags=FD_RAW_INTR;
    cmd.cmd[0]=FD_RECALIBRATE; /* position head to track 0 */
    cmd.cmd[1]=0;
    cmd.cmd_count=2;  /* two bytes in the command */
    if (ioctl(device,FDRAWCMD,&cmd)<0)
      {
         fprintf(stderr,"Error on recalibrate\n");
         exit(1);
       }
  }

void lif_seek(int device, int cylinder)
  {
    struct floppy_raw_cmd cmd;

    cmd.data=NULL;   /* No data to transfer */
    cmd.length=0;
    cmd.rate=RATE250; /* 250kbps data rate */
    cmd.flags=FD_RAW_INTR;
    cmd.cmd[0]=FD_SEEK;  /* Seek to a cylinder */
    cmd.cmd[1]=0;
    cmd.cmd[2]=cylinder;  /* Cylinder to go to */
    cmd.cmd_count=3;
    if (ioctl(device,FDRAWCMD,&cmd)<0)
      {
         fprintf(stderr,"Error on seek to cylinder %d\n",cylinder);
         exit(1);
      }
  }

void lif_read(int device, int cylinder, int head, int sector, 
              unsigned char *data)
  {
    struct floppy_raw_cmd cmd;

    cmd.data=data;  /* Data buffer to transfer */
    cmd.length=SECTOR_SIZE; /* length of data */
    cmd.rate=RATE250; /* 250 kbps */
    cmd.flags=FD_RAW_INTR | FD_RAW_READ; /* Set up DMA, etc */
    cmd.cmd[0]=FD_DD_READ; /* read command */
    cmd.cmd[1]=head?4:0; /* Head select */
    cmd.cmd[2]=cylinder; /* Cylinder value (to check with header) */
    cmd.cmd[3]=head?1:0; /* Head value (to check with header) */
    cmd.cmd[4]=sector; /* Sector to search for */
    cmd.cmd[5]=1; /* 256 byte MFM sectors */
    cmd.cmd[6]=16; /* last sector number on a track */
    cmd.cmd[7]=14; /* gap length */
    cmd.cmd[8]=0xFF; /* 256 bytes, but shouldn't be needed */
    cmd.cmd_count=9;
    if ((ioctl(device,FDRAWCMD,&cmd)<0) || (cmd.reply[0] & 0xC0))
      {
        fprintf(stderr,"Error reading cylinder %d, head %d, sector %d\n",
                cylinder,head,sector);
        exit(1);
      }
  }

void lif_write(int device, int cylinder, int head, int sector, 
               unsigned char *data)
  {
    struct floppy_raw_cmd cmd;

    cmd.data=data;  /* Data buffer to transfer */
    cmd.length=SECTOR_SIZE; /* length of data */
    cmd.rate=RATE250; /* 250 kbps */
    cmd.flags=FD_RAW_INTR | FD_RAW_WRITE; /* Set up DMA, etc */
    cmd.cmd[0]=FD_DD_WRITE; /* write command */
    cmd.cmd[1]=head?4:0; /* Head select */
    cmd.cmd[2]=cylinder; /* Cylinder value (to check with header) */
    cmd.cmd[3]=head?1:0; /* Head value (to check with header) */
    cmd.cmd[4]=sector; /* Sector to search for */
    cmd.cmd[5]=1; /* 256 byte MFM sectors */
    cmd.cmd[6]=16; /* last sector number on a track */
    cmd.cmd[7]=14; /* gap length */
    cmd.cmd[8]=0xFF; /* 256 bytes, but shouldn't be needed */
    cmd.cmd_count=9;
    if ((ioctl(device,FDRAWCMD,&cmd)<0) || (cmd.reply[0] & 0xC0))
      {
        fprintf(stderr,"Error writing cylinder %d, head %d, sector %d\n",
                cylinder,head,sector);
        exit(1);
      }
  }

