/* lifdir.c -- Display the directory of a LIF disk */
/* 2000 A. R. Duell, and placed under the GPL */

#include<stdio.h>
#include<unistd.h>
#include<fcntl.h>
#include"lif_block.h"
#include"lif_dir_utils.h"

#define SECTOR_SIZE 256

/* offsets for the parts of a timestamp */
#define YEAR_OFF 0
#define MONTH_OFF 1
#define DAY_OFF 2
#define HOUR_OFF 3
#define MINUTE_OFF 4
#define SECOND_OFF 5

void print_date(unsigned char *date)
  {
    /* Print an HP-style date and time stamp. On entry, date points to
       the first byte of a 6-byte time stamp */
    printf("%02d/%02d/%02d %02d:%02d:%02d",
         bcd_to_dec(*(date+DAY_OFF)),
         bcd_to_dec(*(date+MONTH_OFF)),
         bcd_to_dec(*(date+YEAR_OFF)),
         bcd_to_dec(*(date+HOUR_OFF)),
         bcd_to_dec(*(date+MINUTE_OFF)),
         bcd_to_dec(*(date+SECOND_OFF)));
  }
       
void print_dir_entry(unsigned char *entry)
  {
    /* Decode and print a directory entry. entry points to a 32 byte 
       directory entry, as described in appendix D of the HP71 HPIL
       owner's manual */
    int i; /* general counter */
    int length; /* File length from directory */
    int total_blocks; /* Number of blocks occupied by file */

    char file_type[10]; /* storage for the file type string */

    /* Print the filename */
    for(i=0; i<10; i++)
      {
        putchar(*(entry+i));
      }
    /* find and print the file type */
    length=file_length(entry,file_type);
    printf("  %-10s  ",file_type);
    total_blocks=get_lif_int(entry+16,4);
    /* Print the length, both from the directory, and from the number of 
       blocks */
    printf("%5d/%-5d     ",length,total_blocks*256);
    /* Print the file time and date */
    if(*(entry+21))
      {
        /* If the month is not 0, it's a valid date */
        print_date(entry+20);
      }
    printf("\n");
  } 

void usage(void)
  {
     fprintf(stderr,"Usage:lifdir [-f] file-or-device\n");
     fprintf(stderr,"      -f flag if directorying image file, not device\n");
     exit(1);
  }

int main(int argc, char **argv)
  {
    /* system variables */
    int option; /* Command line option character */
    int file_flag; /* Input device is a file, not a physical disk */
    int input_device; /* Input file or device descriptor */
    int i; /* General index counter */
    unsigned char data[SECTOR_SIZE]; /* buffer to hold current block */

    /* LIF disk values */
    unsigned int dir_start; /* first block of directory */
    unsigned int dir_length; /* Number of blocks in directory */

    /* Directory search values */
    int dir_end; /* Set at end of directory */
    int dir_entry; /* Directory entry within current block */
    int dir_block; /* Current block offset from start of directory */
    unsigned int file_type; /* File type word */

    /* process command line options */
    file_flag=0; 
    optind=1;
    while((option=getopt(argc,argv,"f?"))!=-1)
      {
        switch(option)
          {
             case 'f' : file_flag=1;
                        break;
             case '?' : usage();
           }
       }

    /* Is one input device specified? */
    if(optind != argc-1)
      {
        /* if not, give error */
        usage();
      }
    
    /* open input device */
    if((input_device=open(argv[argc-1],O_RDONLY,0))==-1)
      {
        fprintf(stderr,"Error opening %s\n",argv[argc-1]);
        exit(1);
      }

    /* if it's a physical device, recalibrate it */
    if(!file_flag)
      {
        lif_init(input_device);
      }

    /* Now read block 0, the volume label block */
    lif_read_block(input_device,file_flag,0,data);

    /* Check that this is a LIF disk or image */
    if(get_lif_int(data+0,2)!=0x8000)
      {
        fprintf(stderr,"This is not a LIF disk!\n");
        exit(1);
      }

    if((*(data+2))!=' ')
      {
        /* There is a volume label */
        printf("Volume : ");
        for(i=2; i<8; i++)
          {
            putchar(*(data+i));
          }
        printf(" ");
       }

     /* If the time stamp month is non-zero, print the time stamp */
    if(*(data+37))
      {
        printf("( formatted : "); 
        print_date(data+36);
        printf(" )");
      }
    printf("\n");

    /* Find where the directory is */
    dir_start=get_lif_int((data+8),4);
    dir_length=get_lif_int((data+16),4);

    /* Now scan the directory */
    for(dir_block=0; dir_block<dir_length; dir_block++)
      {
         dir_end=0;
         lif_read_block(input_device,file_flag,dir_block+dir_start,data);
         for(dir_entry=0; dir_entry<8; dir_entry++)
           {
             file_type=get_lif_int((data+(dir_entry<<5)+10),2);
             if(file_type==0) { continue; } /* Skip over deleted files */
             if(file_type==0xffff)
               {
                 /* This is the end of the directory */
                 dir_end=1;
                 break;
               }
             print_dir_entry(data+(dir_entry<<5));
           }
         if(dir_end) { break; } /* Quit at end of directory */
      }
    close(input_device);
    exit(0); 
  }

