/* lifget.c -- Get a file from a LIF disk */
/* 2000 A. R. Duell, and placed under the GPL */

#include<stdio.h>
#include<unistd.h>
#include<fcntl.h>
#include"lif_block.h"
#include"lif_dir_utils.h"

#define SECTOR_SIZE 256
#define NAME_LEN 10 

void pad_name(unsigned char *name, unsigned char *cmp_name)
  {
    /* Pad a filename with spaces, so as to be able to compare it with
       a LIF directory record. On entry, cmp_name points to a 10 character
       string. On exit, the first characters of this contain the 
       characters from name, the rest are spaces */

    int i; /* general index */  
    for(i=0; i<NAME_LEN; i++)
      {
        cmp_name[i]=' '; /* fill cmp_name with spaces */
      }
    for(i=0; i<NAME_LEN; i++)
      {
        /* Quit at end of name */
        if(!name[i])
          {
            break;
          }
        /* Otherwise copy characters from name to cmp_name */
        cmp_name[i]=name[i];
      }
  }

int compare_names(unsigned char *entry, unsigned char *cmp_name)
  {
    /* Check if the file name in directory entry is the same as the one
       given in cmp_name. Return true if so, false otherwise */

    int i, equal;

    equal=1; /* Initially assume this is the right entry, until proved 
                otherwise */
    for(i=0; i<NAME_LEN; i++)
      {
        if(entry[i]!=cmp_name[i])
          {
            equal=0; /* If the characters don't match, this is not the right
                        entry */
          }
      } 
    return(equal);
  }

void usage(void)
  {
    fprintf(stderr,
    "Usage:lifget [-f] [-d] [-b] file-or-device filename output-file\n");
    fprintf(stderr,"      lifget [-f] [-d] [-b] file-or-device filename\n");
    fprintf(stderr,"      (Output goes to standard output)\n");
    fprintf(stderr,"\n");
    fprintf(stderr,
    "      -f flag if getting from an image file, not a device\n");
    fprintf(stderr,
    "      -d flag to include directory entry on start of file \n"); 
    fprintf(stderr,"      -b flag to copy the blocks used by the file,\n");
    fprintf(stderr,"          ignoring the file length information\n");
    exit(1);
  }

void file_copy(int input_device, FILE* output_file, int file_flag, 
               int start, int length)
  {
    /* Copy length bytes starting at block start from LIF input_device
       to output_file */
  
    unsigned int complete_blocks; /* Number of complete blocks to copy */
    unsigned int leftover_bytes; /* And the odd bytes on the end */
    unsigned int block; /* Current block offset from start */
    unsigned int byte; /* Current byte offset */
    unsigned char data[SECTOR_SIZE];

    if(complete_blocks=length/SECTOR_SIZE)
      {
        /* Copy the complete blocks first */
        for(block=0; block<complete_blocks; block++)
          {
            lif_read_block(input_device,file_flag,start+block,data);
            fwrite(data,sizeof(char),SECTOR_SIZE,output_file);
          }
      }
    if(leftover_bytes=length%SECTOR_SIZE)
      {
        /* Odd bytes on the end -- read one more block */
        lif_read_block(input_device,file_flag,start+complete_blocks,data);
        fwrite(data,sizeof(char),leftover_bytes,output_file);
      }
  }

int main(int argc, char **argv)
  {
    /* System variables */
    int option; /* Command line option character */
    int file_flag; /* Input device is a file, not a physical disk */
    int dir_flag; /* Include directory entry on start of output */
    int block_flag; /* Copy blocks */
    int input_device; /* Descriptor of input device */
    FILE *output_file; /* Output file stream */
    unsigned char cmp_name[10]; /* File name to look for */
    
    /* LIF disk values */
    unsigned int dir_start; /* first block of the directory */
    unsigned int dir_length; /* length of directory in blocks */

    /* Directory search values */
    int dir_end; /* Set at end of directory */
    int found_file; /* Set when file found */
    int dir_entry; /* Directory entry within current block */
    int dir_block; /* Current block offset from start of directory */
    unsigned int file_type; /* file type word */
    unsigned char dir_data[SECTOR_SIZE]; /* Current directory block data */

    /* File values */
    int file_start; /* Starting block number of the file to copy */
    int file_len; /* Length of file in bytes */

    /* Process command line options */
    file_flag=0;
    dir_flag=0;
    block_flag=0;
    optind=1;
    while ((option=getopt(argc,argv,"fdb?"))!=-1)
      {
        switch(option)
          {
            case 'f' : file_flag=1;
                       break;
            case 'd' : dir_flag=1;
                       break;
            case 'b' : block_flag=1;
                       break;
            case '?' : usage();
          }
      }

    /* Are the right number of names specified ? */
    if( (optind != argc-2) && (optind != argc-3) )
      {
        /* No, give an error */
        usage();
      }

    /* Open input device */
    if((input_device=open(argv[optind],O_RDONLY,0))==-1)
      {
        fprintf(stderr,"Error opening %s\n",argv[optind]);
        exit(1);
      }

    /* If it's a physical device, recalibrate it */
    if(!file_flag)
      {
        lif_init(input_device);
      }

    /* Now read block 0 to find where the directory is */
    lif_read_block(input_device,file_flag,0,dir_data);

    /* Make sure it's a LIF disk */
    if(get_lif_int(dir_data+0,2)!=0x8000)
      {
        fprintf(stderr,"This is not a LIF disk!\n");
        exit(1);
      }

    /* Find the directory */
    dir_start=get_lif_int(dir_data+8,4);
    dir_length=get_lif_int(dir_data+16,4);

    /* Pad the filename with spaces to enable comparison */
    pad_name(argv[optind+1],cmp_name);

    /* Scan the directory */
    dir_end=0;
    found_file=0;
    for(dir_block=0; dir_block<dir_length; dir_block++)
      {
        lif_read_block(input_device,file_flag,dir_block+dir_start,dir_data);
        for(dir_entry=0; dir_entry<8; dir_entry++)
          {
            file_type=get_lif_int(dir_data+(dir_entry<<5)+10,2);
            if(file_type==0) { continue; } /* Skip deleted files */ 
            if(file_type==0xFFFF)
              {
                /* End of directory */
                dir_end=1;
                break;
              }
            if(compare_names(dir_data+(dir_entry<<5),cmp_name))
              {
                /* Found the file */
                found_file=1;
                break;
              }
          }
        if(dir_end || found_file) { break; }; /* Quit at end or if file found */
      }

    if(!found_file)
      {
        /* Give file not found error */
        fprintf(stderr,"File %s not found\n",argv[optind+1]);
        exit(2);
      }

    /* Open the output file */
    if(optind==argc-3)
      {
        output_file=fopen(argv[argc-1],"w");
      }
    else
      {
        output_file=stdout;
      }

    /* If the -d flag was specified, send the directory entry */
    if(dir_flag)
      {
        fwrite(dir_data+(dir_entry<<5),sizeof(char),32,output_file);
      }

    /* Find the file start and length */
    file_start=get_lif_int(dir_data+(dir_entry<<5)+12,4);
    if(block_flag)
      {
        file_len=get_lif_int(dir_data+(dir_entry<<5)+16,4)*256;
      }
    else
      {
        file_len=file_length(dir_data+(dir_entry<<5),NULL);
      }

    /* Actually copy the file */ 
    file_copy(input_device,output_file,file_flag,file_start,file_len);

    /* tidy up and quit */
    if(optind==argc-3)
      {
        fclose(output_file);
      }
    close(input_device);
    exit(0);      
  }
