/* lifstat.c -- Display status information of a LIF disk */
/* 2001 A. R. Duell, and placed under the GPL */

#include<stdio.h>
#include<stdlib.h>
#include<unistd.h>
#include<fcntl.h>
#include"lif_block.h"
#include"lif_dir_utils.h"

/* parameters of the LIF disk */
#define SECTOR_SIZE 256
#define SPT 16
#define HEADS 2
#define CYLINDERS 77
#define TOTAL_BLOCKS 2464


void print_block_no(unsigned int block_no)
/* Print the block number, and it's address (cylinder/head/sector) to 
   standard output */
  {
    printf("%d (%d/%d/%d)",
            block_no,
            (int) block_no/(HEADS*SPT),
            (int) (block_no/SPT)%HEADS,
            (block_no%SPT)+1);
  }

void lif_status(int input_file, int file_flag, int block_flag, unsigned 
                int block_no)
/* Print various status items for the given LIF disk (or image) to standard
   output. If block_flag is false, then a sumary of data for the entire disk
   is produced. If block_flag is true, then the name of the file 
   containing block_no is output */
  {
    unsigned int dir_start; /* first directory block */
    unsigned int dir_length; /* size of directory */
    int dir_end; /* Set at end of directory search */
    int found; /* Set if given block found */
    int dir_entry; /* directory entry in current block */
    int dir_block; /* current directory block (offset from dir_start) */
    unsigned int file_type; /* current file type */
    unsigned int file_start; /* start block of current file */
    unsigned int file_length; /* length of current file */
    unsigned int last_block; /* last used block on the disk */
    int i; /* loop index */
    char c; /* filename character */
    unsigned char data[SECTOR_SIZE]; /* disk block */

    /* Handle the fixed cases */
    if((block_flag)&&(block_no==0))
      {
        print_block_no(block_no);
        printf(" : Volume Label\n");
        return;
      }
    if((block_flag)&&(block_no==1))
      {
        print_block_no(block_no);
        printf(" : System 3000 block\n");
        return;
      }

    /* It's necessary to read the volume label block */
    lif_read_block(input_file,file_flag,0,data);
    /* Check it's a LIF disk */
    if(get_lif_int(data,2)!=0x8000)
      {
        fprintf(stderr,"This is not a LIF disk!\n");
        exit(1);
      }
    /* Find the directory */
    dir_start=get_lif_int(data+8,4);
    dir_length=get_lif_int(data+16,4);

    if(block_flag)
      {
        /* looking for a particular block, is it in the directory ? */
        if((block_no>=dir_start)&&(block_no<dir_start+dir_length))
          {
            print_block_no(block_no);
            printf(" : directory\n");
            return;
          }
      }
    else
      {
        /* If not looking for a particular block, print directory location */
        printf("Directory start : ");
        print_block_no(dir_start);
        printf(" end : ");
        print_block_no(dir_start+dir_length-1);
        printf("\n");
      }

    /* Scan the directory */
    found=0;
    dir_end=0;
    for(dir_block=0; (dir_block<dir_length)&&(!dir_end)&&(!found); dir_block++)
      {
        lif_read_block(input_file,file_flag,dir_block+dir_start,data);
        for(dir_entry=0; dir_entry<8; dir_entry++)
          {
            file_type=get_lif_int(data+(dir_entry<<5)+10,2);
            if(file_type==0) { continue; } /* skip deleted files */
            if(file_type==0xffff)
              {
                /* End of directory */
                dir_end=1;
                break;
              }
            file_start=get_lif_int(data+(dir_entry<<5)+12,4);
            file_length=get_lif_int(data+(dir_entry<<5)+16,4);
            /* update last used block */
            last_block=file_start+file_length-1;
            /* If searching for a particular block, is it in this file ? */
            if(block_flag && (block_no>=file_start) && 
               (block_no<file_start+file_length))
              {
                found=1;
                print_block_no(block_no);
                printf(" : ");
                for(i=0; i<10; i++)
                  {
                    if((c=*(data+(dir_entry<<5)+i))==' ') { break; }
                    putchar(c);
                  }
                printf("\n");
              }
          }
      }

    if(block_flag && !found)
      {
        /* If the block wasn't part of a file, it's unused */
        print_block_no(block_no);
        printf(" : unused\n");
      }

    if(!block_flag)
      {
        printf("User data start : ");
        print_block_no(dir_start+dir_length);
        printf(" end : ");
        print_block_no(last_block);
        printf("\n");
      }
  }

void usage(void)
  {
    fprintf(stderr,"Usage : lifstat [-f] input_file_or_device\n");
    fprintf(stderr,"        print summary of LIF disk stats\n\n");
    fprintf(stderr,"        lifstat [-f] input_file_or_device block_no\n");
    fprintf(stderr,
      "        lifstat [-f] input_file_or_device cylinder head sector \n");
    fprintf(stderr,"        print filename containing given block\n");
    exit(1);
  }

int main(int argc, char **argv)
  {
    int option; /* command line option character */
    int file_flag; /* input comes from image file, not device */
    int input_device; /* input file or device desciptor */
    unsigned int block_no; /* desired block number */
    unsigned int cylinder,head,sector; /* desired disk address */

    file_flag=0;
    optind=1;
    /* Decode command line options */
    while((option=getopt(argc,argv,"f?"))!=-1)
      {
        switch(option)
          {
            case 'f' : file_flag=1;
                       break;
            case '?' : usage();
          }
      }

    /* Is there a file name ? */
    if(optind==argc)
      {
        /* No, give error */
        usage();
      }

    /* Open input device */
    if((input_device=open(argv[optind],O_RDONLY,0))==-1)
      {
        fprintf(stderr,"Error opening %s\n",argv[optind]);
        exit(1);
      }
    if(!file_flag)
      {
        lif_init(input_device);
      }

    /* process other command line arguments */
    optind++;
    switch(argc-optind)
      {
        case 0 : lif_status(input_device,file_flag,0,0);
                 break;
        case 1 : block_no=atoi(argv[optind]);
                 if(block_no>=TOTAL_BLOCKS)
                   {
                     fprintf(stderr,"Block number out of range\n");
                     exit(1);
                   }
                 lif_status(input_device,file_flag,1,block_no);
                 break;
        case 3 : cylinder = atoi(argv[optind++]);
                 head = atoi(argv[optind++]);
                 sector = atoi(argv[optind])-1;
                 if(cylinder>=CYLINDERS)
                   {
                     fprintf(stderr,"Cylinder out of range\n");
                     exit(1);
                   }
                 if(head>=HEADS)
                   {
                     fprintf(stderr,"Head out of range\n");
                     exit(1);
                   }
                 if(sector>=SPT)
                   {
                     fprintf(stderr,"Sector out of range\n");
                     exit(1);
                   }
                 block_no=cylinder*HEADS*SPT + head*SPT + sector;
                 lif_status(input_device,file_flag,1,block_no);
                 break;
        default : usage();
      }
    close(input_device);
    exit(0);
  }

